import typing

import numpy as np
import torch
import pytorch_kinematics as pk
from pytorch_volumetric import sdf
import logging

logger = logging.getLogger(__file__)


class RobotSDF(sdf.ObjectFrameSDF):
    """Create an SDF for a robot model described by a pytorch_kinematics Chain.
    The SDF is conditioned on a joint configuration which must be set."""

    def __init__(self, chain: pk.Chain, default_joint_config=None, path_prefix='',
                 link_sdf_cls: typing.Callable[[sdf.ObjectFactory], sdf.ObjectFrameSDF] = sdf.MeshSDF):
        """

        :param chain: Robot description; each link should be a mesh type - non-mesh geometries are ignored
        :param default_joint_config: values for each joint of the robot by default; None results in all zeros
        :param path_prefix: path to search for referenced meshes inside the robot description (e.g. URDF) which may use
        relative paths. This given path is prefixed onto those relative paths in order to find the meshes.
        :param link_sdf_cls: Factory of each link's SDFs; **kwargs are forwarded to this factory
        :param kwargs: Keyword arguments fed to link_sdf_cls
        """
        self.chain = chain
        self.dtype = self.chain.dtype
        self.device = self.chain.device
        self.q = None
        self.object_to_link_frames: typing.Optional[pk.Transform3d] = None
        self.joint_names = self.chain.get_joint_parameter_names()
        self.frame_names = self.chain.get_frame_names(exclude_fixed=False)
        self.sdf: typing.Optional[sdf.ComposedSDF] = None
        self.sdf_to_link_name = []
        self.configuration_batch = None

        sdfs = []
        offsets = []
        # get the link meshes from the frames and create meshes
        for frame_name in self.frame_names:
            frame = self.chain.find_frame(frame_name)
            # TODO create SDF for non-mesh primitives
            # TODO consider the visual offset transform
            for link_vis in frame.link.visuals:
                if link_vis.geom_type == "mesh":
                    logger.info(f"{frame.link.name} offset {link_vis.offset}")
                    link_obj = sdf.MeshObjectFactory(link_vis.geom_param, path_prefix=path_prefix)
                    link_sdf = link_sdf_cls(link_obj)
                    self.sdf_to_link_name.append(frame.link.name)
                    sdfs.append(link_sdf)
                    offsets.append(link_vis.offset)
                else:
                    logger.warning(f"Cannot handle non-mesh link visual type {link_vis}")

        self.offset_transforms = offsets[0].stack(*offsets[1:]).to(device=self.device, dtype=self.dtype)
        self.sdf = sdf.ComposedSDF(sdfs, self.object_to_link_frames)
        self.set_joint_configuration(default_joint_config)

    def surface_bounding_box(self, padding=0.):
        return self.sdf.surface_bounding_box(padding=padding)

    def link_bounding_boxes(self):
        """
        Get the bounding box of each link in the robot's frame under the current configuration.
        Note that the bounding box is not necessarily axis-aligned, so the returned bounding box is not just
        the min and max of the points.
        :return: [A x] [B x] 8 x 3 points of the bounding box for each link in the robot's frame
        """
        tfs = self.sdf.obj_frame_to_link_frame.inverse()
        bbs = []
        for i in range(len(self.sdf.sdfs)):
            sdf = self.sdf.sdfs[i]
            bb = aabb_to_ordered_end_points(sdf.surface_bounding_box(padding=0))
            bb = tfs.transform_points(torch.tensor(bb, device=tfs.device, dtype=tfs.dtype))[
                self.sdf.ith_transform_slice(i)]
            bbs.append(bb)
        return torch.stack(bbs).squeeze()

    def set_joint_configuration(self, joint_config=None):
        """
        Set the joint configuration of the robot
        :param joint_config: [A x] M optionally arbitrarily batched joint configurations. There are M joints; A can be
        any number of batched dimensions.
        :return:
        """
        M = len(self.joint_names)
        if joint_config is None:
            joint_config = torch.zeros(M, device=self.device, dtype=self.dtype)
        # Transform3D only works with 1 batch dimension, so we need to manually flatten any additional ones
        # save the batch dimensions for when retrieving points
        if len(joint_config.shape) > 1:
            self.configuration_batch = joint_config.shape[:-1]
            joint_config = joint_config.reshape(-1, M)
        else:
            self.configuration_batch = None
        tf = self.chain.forward_kinematics(joint_config, end_only=False)
        tsfs = []
        for link_name in self.sdf_to_link_name:
            tsfs.append(tf[link_name].get_matrix())
        # make offset transforms have compatible batch dimensions
        offset_tsf = self.offset_transforms
        if self.configuration_batch is not None:
            offset_tsf = pk.Transform3d(matrix=offset_tsf.get_matrix().repeat(*self.configuration_batch, 1, 1))
        self.object_to_link_frames = pk.Transform3d(matrix=torch.cat(tsfs).inverse()).compose(offset_tsf)
        if self.sdf is not None:
            self.sdf.set_transforms(self.object_to_link_frames, batch_dim=self.configuration_batch)

    def __call__(self, points_in_object_frame):
        """
        Query for SDF value and SDF gradients for points in the robot's frame
        :param points_in_object_frame: [B x] N x 3 optionally arbitrarily batched points in the robot frame; B can be
        any number of batch dimensions.
        :return: [A x] [B x] N SDF value, and [A x] [B x] N x 3 SDF gradient. A are the configurations' arbitrary
        number of batch dimensions.
        """
        return self.sdf(points_in_object_frame)


def cache_link_sdf_factory(resolution=0.01, padding=0.1, **kwargs):
    def create_sdf(obj_factory: sdf.ObjectFactory):
        gt_sdf = sdf.MeshSDF(obj_factory)
        return sdf.CachedSDF(obj_factory.name, resolution, obj_factory.bounding_box(padding=padding), gt_sdf, **kwargs)

    return create_sdf


def aabb_to_ordered_end_points(aabb):
    aabbMin = aabb[:, 0]
    aabbMax = aabb[:, 1]
    arr = [
        [aabbMin[0], aabbMin[1], aabbMin[2]],
        [aabbMax[0], aabbMin[1], aabbMin[2]],
        [aabbMin[0], aabbMax[1], aabbMin[2]],
        [aabbMin[0], aabbMin[1], aabbMax[2]],
        [aabbMin[0], aabbMax[1], aabbMax[2]],
        [aabbMax[0], aabbMin[1], aabbMax[2]],
        [aabbMax[0], aabbMax[1], aabbMin[2]],
        [aabbMax[0], aabbMax[1], aabbMax[2]]
    ]
    if torch.is_tensor(aabb):
        return torch.tensor(arr, device=aabb.device, dtype=aabb.dtype)
    return np.array(arr)



import os
import math
import torch
import time
from matplotlib import pyplot as plt
import numpy as np
from timeit import default_timer as timer
import open3d as o3d
import pytorch_kinematics as pk
import pytorch_volumetric as pv

import pybullet as p
import pybullet_data

import logging

plt.switch_backend('Qt5Agg')

logger = logging.getLogger(__file__)
logging.basicConfig(level=logging.INFO, force=True,
                    format='[%(levelname)s %(asctime)s %(pathname)s:%(lineno)d] %(message)s',
                    datefmt='%m-%d %H:%M:%S')

TEST_DIR = os.path.dirname(__file__)


def test_urdf_to_sdf():
    visualization = "open3d"
    urdf = "kuka_iiwa/model.urdf"
    search_path = pybullet_data.getDataPath()
    full_urdf = os.path.join(search_path, urdf)
    chain = pk.build_serial_chain_from_urdf(open(full_urdf).read(), "lbr_iiwa_link_7")
    d = "cuda" if torch.cuda.is_available() else "cpu"

    chain = chain.to(device=d)
    # use MeshSDF or CachedSDF for much faster lookup
    s = pv.RobotSDF(chain, path_prefix=os.path.join(search_path, "kuka_iiwa"), )
    # link_sdf_cls=pv.cache_link_sdf_factory(resolution=0.02, padding=0.1, device=d))
    th = torch.tensor([0.0, -math.pi / 4.0, 0.0, math.pi / 2.0, 0.0, math.pi / 4.0, 0.0], device=d)

    s.set_joint_configuration(th)

    y = 0.02
    query_range = np.array([
        [-1, 0.5],
        [y, y],
        [-0.2, 0.8],
    ])

    plt.ion()
    plt.show()

    ret = pv.draw_sdf_slice(s, query_range, resolution=0.01, device=s.device)
    sdf_val = ret[0]
    pts = ret[2]

    surface = sdf_val.abs() < 0.005

    if visualization == "pybullet":
        # toggles - g:GUI w:wireframe j:joint axis a:AABB i:interrupt
        p.connect(p.GUI)
        p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
        p.setAdditionalSearchPath(search_path)
        armId = p.loadURDF(urdf, [0, 0, 0], useFixedBase=True)
        # p.resetBasePositionAndOrientation(armId, [0, 0, 0], [0, 0, 0, 1])
        for i, q in enumerate(th):
            p.resetJointState(armId, i, q.item())

        try:
            from base_experiments.env.env import draw_ordered_end_points
            from base_experiments.env.pybullet_env import DebugDrawer
            vis = DebugDrawer(1., 1.5)
            vis.toggle_3d(True)
            vis.set_camera_position([-0.1, 0, 0], yaw=-30, pitch=-20)
            # draw bounding box for each link (set breakpoints here to better see the link frame bounding box)
            tfs = s.sdf.obj_frame_to_link_frame.inverse()
            for i in range(len(th)):
                sdf = s.sdf.sdfs[i]
                aabb = pv.aabb_to_ordered_end_points(np.array(sdf.ranges))
                aabb = tfs.transform_points(torch.tensor(aabb, device=tfs.device, dtype=tfs.dtype))[i]
                draw_ordered_end_points(vis, aabb)
                time.sleep(0.2)

            vis.draw_points("surface", pts[surface])
        except:
            pass
        finally:
            p.disconnect()
    elif visualization == "open3d":
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(pts[surface].cpu().numpy())
        o3d.visualization.draw_geometries(pv.get_transformed_meshes(s) + [pcd])


def test_batch_over_configurations():
    urdf = "kuka_iiwa/model.urdf"
    search_path = pybullet_data.getDataPath()
    full_urdf = os.path.join(search_path, urdf)
    chain = pk.build_serial_chain_from_urdf(open(full_urdf).read(), "lbr_iiwa_link_7")
    d = "cuda" if torch.cuda.is_available() else "cpu"

    chain = chain.to(device=d)
    s = pv.RobotSDF(chain, path_prefix=os.path.join(search_path, "kuka_iiwa"),
                    link_sdf_cls=pv.cache_link_sdf_factory(resolution=0.02, padding=1.0, device=d))

    th = torch.tensor([0.0, -math.pi / 4.0, 0.0, math.pi / 2.0, 0.0, math.pi / 4.0, 0.0], device=d)
    N = 20
    th_perturbation = torch.randn(N - 1, 7, device=d) * 0.1
    th = torch.cat((th.view(1, -1), th_perturbation + th))

    s.set_joint_configuration(th)

    y = 0.02
    query_range = np.array([
        [-1, 0.5],
        [y, y],
        [-0.2, 0.8],
    ])

    coords, pts = pv.get_coordinates_and_points_in_grid(0.01, query_range, device=s.device)

    start = timer()
    all_sdf_val, all_sdf_grad = s(pts)
    elapsed = timer() - start
    logger.info("configurations: %d points: %d elapsed: %fms time per config and point: %fms", N, len(pts),
                elapsed * 1000, elapsed * 1000 / N / len(pts))

    for i in range(N):
        th_i = th[i]
        s.set_joint_configuration(th_i)
        sdf_val, sdf_grad = s(pts)

        assert torch.allclose(sdf_val, all_sdf_val[i])
        assert torch.allclose(sdf_grad, all_sdf_grad[i], atol=1e-6)


def test_bounding_box():
    urdf = "kuka_iiwa/model.urdf"
    search_path = pybullet_data.getDataPath()
    full_urdf = os.path.join(search_path, urdf)
    chain = pk.build_serial_chain_from_urdf(open(full_urdf).read(), "lbr_iiwa_link_7")
    d = "cuda" if torch.cuda.is_available() else "cpu"

    chain = chain.to(device=d)
    # use MeshSDF or CachedSDF for much faster lookup
    s = pv.RobotSDF(chain, path_prefix=os.path.join(search_path, "kuka_iiwa"), )
    th = torch.tensor([0.0, -math.pi / 4.0, 0.0, math.pi / 2.0, 0.0, math.pi / 4.0, 0.0], device=d)

    s.set_joint_configuration(th)

    # toggles - g:GUI w:wireframe j:joint axis a:AABB i:interrupt
    p.connect(p.GUI)
    p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
    p.setAdditionalSearchPath(search_path)
    armId = p.loadURDF(urdf, [0, 0, 0], useFixedBase=True)
    # p.resetBasePositionAndOrientation(armId, [0, 0, 0], [0, 0, 0, 1])
    for i, q in enumerate(th):
        p.resetJointState(armId, i, q.item())

    try:
        from base_experiments.env.env import draw_ordered_end_points, draw_AABB
        from base_experiments.env.pybullet_env import DebugDrawer
        delay = 0.2
        vis = DebugDrawer(1., 1.5)
        vis.toggle_3d(True)
        vis.set_camera_position([-0.1, 0, 0], yaw=-30, pitch=-20)
        # draw bounding box for each link (set breakpoints here to better see the link frame bounding box)
        bbs = s.link_bounding_boxes()
        for i in range(len(s.sdf.sdfs)):
            bb = bbs[i]
            draw_ordered_end_points(vis, bb)
            time.sleep(delay)
        # total aabb
        aabb = s.surface_bounding_box(padding=0)
        draw_AABB(vis, aabb.cpu().numpy())
        time.sleep(delay)
    except ImportError as e:
        print(e)

    time.sleep(2)
    p.disconnect()


def test_single_link_robot():
    full_urdf = os.path.join(TEST_DIR, 'offset_wrench.urdf')
    chain = pk.build_serial_chain_from_urdf(open(full_urdf).read(), "offset_wrench")
    d = "cuda" if torch.cuda.is_available() else "cpu"

    chain = chain.to(device=d)
    # paths to the link meshes are specified with their relative path inside the URDF
    # we need to give them the path prefix as we need their absolute path to load
    sdf = pv.RobotSDF(chain, path_prefix=TEST_DIR,
                      link_sdf_cls=pv.cache_link_sdf_factory(resolution=0.001, padding=0.05, device=d))
    trans_x, trans_y, trans_z = 0.0, 0.0, 0.0
    rot_x, rot_y, rot_z = 0.0, 0.0, 0.0
    trans = torch.tensor([trans_x, trans_y, trans_z], device=d)
    rot = torch.tensor([rot_x, rot_y, rot_z], device=d)
    H = torch.eye(4, device=d)
    H[:-1, -1] = trans
    H[:-1, :-1] = pk.euler_angles_to_matrix(rot, 'XYZ')

    th = torch.cat((trans, rot), dim=0)
    sdf.set_joint_configuration(th.view(1, -1))
    query_range = sdf.surface_bounding_box(padding=0.05)
    # M x 3 points
    coords, pts = pv.get_coordinates_and_points_in_grid(0.001, query_range, device=sdf.device)

    sdf_val, sdf_grad = sdf(pts)
    # because we passed in th with size (1, 6), the output is also (1, M) and (1, M, 3) meaning we have a batch of 1
    sdf_val = sdf_val[0]
    sdf_grad = sdf_grad[0]
    near_surface = sdf_val.abs() < 0.001
    surf_pts = pts[near_surface]
    surf_norms = sdf_grad[near_surface]
    pc = o3d.geometry.PointCloud()
    pc.points = o3d.utility.Vector3dVector(surf_pts.cpu())
    pc.normals = o3d.utility.Vector3dVector(surf_norms.cpu())
    o3d.visualization.draw_geometries([pc])


if __name__ == "__main__":
    test_urdf_to_sdf()
    test_batch_over_configurations()
    test_bounding_box()
    test_single_link_robot()



from pytorch_volumetric.chamfer import batch_chamfer_dist
from pytorch_volumetric.sdf import sample_mesh_points, ObjectFrameSDF, MeshSDF, CachedSDF, ComposedSDF, SDFQuery, \
    ObjectFactory, MeshObjectFactory
from pytorch_volumetric.voxel import Voxels, VoxelGrid, VoxelSet, ExpandingVoxelGrid, get_divisible_range_by_resolution, \
    get_coordinates_and_points_in_grid
from pytorch_volumetric.model_to_sdf import RobotSDF, cache_link_sdf_factory, aabb_to_ordered_end_points
from pytorch_volumetric.visualization import draw_sdf_slice, get_transformed_meshes



import copy

from pytorch_volumetric import voxel
from pytorch_volumetric import sdf
from pytorch_volumetric import model_to_sdf
from matplotlib import pyplot as plt
import matplotlib.colors


def fmt(x):
    s = f"{x:.1f}"
    if s.endswith("0"):
        s = f"{x:.0f}"
    if x == 0:
        return "surface"
    return rf"{s}" if plt.rcParams["text.usetex"] else f"{s}"


def draw_sdf_slice(s: sdf.ObjectFrameSDF, query_range, resolution=0.01, interior_padding=0.2,
                   cmap="Greys_r", device="cpu"):
    """

    :param s: SDF to query on
    :param query_range: (min, max) for each dimension x,y,z. One dimension must have min=max to be sliced along, with
    the other dimensions shown. Note that this should be given in the SDF's frame.
    :param resolution:
    :param interior_padding:
    :param cmap: matplotlib compatible colormap
    :param device: pytorch device
    :return:
    """
    coords, pts = voxel.get_coordinates_and_points_in_grid(resolution, query_range, device=device)
    dim_labels = ['x', 'y', 'z']
    slice_dim = None
    for i in range(len(dim_labels)):
        if len(coords[i]) == 1:
            slice_dim = i
            break

    # to properly draw a slice, the coords for that dimension must have only 1 element
    if slice_dim is None:
        raise RuntimeError(f"Sliced SDF requires a single query value for the sliced, but all query dimensions > 1")

    shown_dims = [i for i in range(3) if i != slice_dim]

    sdf_val, sdf_grad = s(pts)
    norm = matplotlib.colors.Normalize(vmin=sdf_val.min().cpu() - interior_padding, vmax=sdf_val.max().cpu())

    ax = plt.gca()
    ax.set_xlabel(dim_labels[shown_dims[0]])
    ax.set_ylabel(dim_labels[shown_dims[1]])
    x = coords[shown_dims[0]].cpu()
    z = coords[shown_dims[1]].cpu()
    v = sdf_val.reshape(len(x), len(z)).transpose(0, 1).cpu()
    cset1 = ax.contourf(x, z, v, norm=norm, cmap=cmap)
    cset2 = ax.contour(x, z, v, colors='k', levels=[0], linestyles='dashed')
    ax.clabel(cset2, cset2.levels, inline=True, fontsize=13, fmt=fmt)
    # fig = plt.gcf()
    # fig.canvas.draw()
    plt.draw()
    plt.pause(0.005)
    return sdf_val, sdf_grad, pts, ax, cset1, cset2


def get_transformed_meshes(robot_sdf: model_to_sdf.RobotSDF, obj_to_world_tsf=None):
    """Get the meshes of each link of the robot, transformed to the world frame.
    Each link is assumed to be a MeshSDF.
    You can use this like:

    import open3d as o3d
    meshes = get_transformed_meshes(robot_sdf)
    o3d.visualization.draw_geometries(meshes)
    """

    meshes = []
    # link to obj in the form of (object) H (link)
    tsfs = robot_sdf.sdf.obj_frame_to_link_frame.inverse()
    # given a transform in the form of (world) H (object)
    if obj_to_world_tsf is not None:
        # compose the transform to get (world) H (link)
        tsfs = obj_to_world_tsf.compose(tsfs)
    tsfs = tsfs.get_matrix()
    for i in range(len(robot_sdf.sdf_to_link_name)):
        # assuming they are individually MeshSDFs
        mesh = copy.deepcopy(robot_sdf.sdf.sdfs[i].obj_factory._mesh)
        mesh = mesh.transform(tsfs[i].cpu().numpy())
        meshes.append(mesh)
    return meshes



import open3d as o3d
import torch
import os
import pytorch_volumetric as pv
from pytorch_volumetric import sample_mesh_points

TEST_DIR = os.path.dirname(__file__)


def test_gradients_at_surface_pts(mesh):
    d = "cuda" if torch.cuda.is_available() else "cpu"

    # press n to visualize the normals / gradients
    obj = pv.MeshObjectFactory(os.path.join(TEST_DIR, mesh))
    sdf = pv.MeshSDF(obj)

    # sample points on the obj mesh surface uniformly
    pts, normals, _ = sample_mesh_points(obj, name=mesh, num_points=1000)

    # query the sdf value and gradient at the sampled points
    sdf_vals, sdf_grads = sdf(pts)

    assert torch.allclose(sdf_vals.abs(), torch.zeros_like(sdf_vals), atol=1e-4)

    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(pts.cpu())
    pcd.normals = o3d.utility.Vector3dVector(normals.cpu())
    pcd.paint_uniform_color([0, 1, 0])
    o3d.visualization.draw_geometries([sdf.obj_factory._mesh, pcd])

    # compare the gradient against the surface normals plotted before - should be the same
    pcd.normals = o3d.utility.Vector3dVector(sdf_grads.cpu())
    # color based on sdf value
    colors = torch.zeros_like(sdf_vals).view(-1, 1).repeat(1, 3)
    colors[:, 0] = (sdf_vals - sdf_vals.min()) / (sdf_vals.max() - sdf_vals.min())
    pcd.colors = o3d.utility.Vector3dVector(colors.cpu())
    o3d.visualization.draw_geometries([sdf.obj_factory._mesh, pcd])

    # sample points in the bounding box
    coords, pts = pv.get_coordinates_and_points_in_grid(0.002, obj.bounding_box(0.01), device=d)
    # randomly downsample to some number of points
    pts = pts[torch.randperm(len(pts))[:1000]]
    # query the sdf value and gradient at the sampled points
    sdf_vals, sdf_grads = sdf(pts)
    # visualize the sdf value and gradient at the sampled points
    pcd.points = o3d.utility.Vector3dVector(pts.cpu())
    pcd.normals = o3d.utility.Vector3dVector(sdf_grads.cpu())
    colors = torch.zeros_like(sdf_vals).view(-1, 1).repeat(1, 3)
    colors[:, 0] = (sdf_vals - sdf_vals.min()) / (sdf_vals.max() - sdf_vals.min())
    colors[:, 1] = 1
    pcd.colors = o3d.utility.Vector3dVector(colors.cpu())
    o3d.visualization.draw_geometries([sdf.obj_factory._mesh, pcd])


if __name__ == "__main__":
    test_gradients_at_surface_pts("probe.obj")
    test_gradients_at_surface_pts("offset_wrench_nogrip.obj")
